
# THIS FILE IS GENERATED FROM CVXPY SETUP.PY
short_version = '1.8.2'
version = '1.8.2'
full_version = '1.8.2'
git_revision = '76b0527'
commit_count = '255'
release = True
if not release:
    version = full_version
